# ----------------------------------------------------------------------------------------------------
# Provider para obtenção de datasets persistidos na área de armazenamento local
# ----------------------------------------------------------------------------------------------------
import os
from io import BytesIO
from ..utils import load_env_variables, get_file_local, make_log
from pathlib import Path

# Para facilitar, define um logger único para todas as funções
LOGGER = make_log("LOG_MLLIB.log")


def load_datasets_local(datasets_filenames: dict) -> dict:
    """
    Carrega os datasets necessários para o modelo que foram persistidos na área de armazenamento local. Os parâmetros
    de acesso deverão ser fornecidos através de um arquivo chamado '.env' que deve ser criado no repositório local e
    preenchido com a seguinte variável: LOCAL_PATH = "caminho local onde os datasets se encontram". Dica de
    segurança: Não deixe o arquivo '.env' ser versionado/persistido no repositório remoto do código.
        :param datasets_filenames: Dicionário contendo os tipos de datasets e os nomes dos respectivos arquivos.
                                   Exemplo: {'features': 'nome_arquivo_features', 'targets': 'nome_arquivo_targets'}
        :return: Dicionário com os datasets carregados.
    """
    # Obtém as informações necessárias para carregar os arquivos
    load_env_variables()
    local_path = os.environ.get("LOCAL_PATH")

    if local_path is None:
        msg = f"Não foram encontradas todas as variáveis de ambiente necessárias. Certifique-se que um arquivo " \
              f"chamado '.env' exista; esteja localizado na pasta da aplicação e que possua valor para a variável: " \
              f"'LOCAL_PATH'. Ou se preferir, configure essa variável de ambiente e seu respectivo valor. " \
              f"Programa abortado!"
        LOGGER.error(msg)
        raise RuntimeError(msg)

    datasets = {}

    for tipo, nome_arquivo in datasets_filenames.items():
        datasets[tipo] = BytesIO(get_file_local(str(Path(local_path) / nome_arquivo)))

    return datasets



# ----------------------------------------------------------------------------------------------------
# Funções úteis que poderão ser utilizadas em qualquer parte do código.
# ----------------------------------------------------------------------------------------------------
import minio
import logging
import configparser
from pathlib import Path
from os import makedirs
from dotenv import load_dotenv, find_dotenv


def make_log(filename: str) -> logging.Logger:
    """
    Cria um arquivo para geração de logs. Se o mesmo já existir, inicia a gravação a partir do final desse arquivo.
    Também retorna um logger para gravação dos logs.
        :param filename: Nome do arquivo de logs.
        :return: Um logger para gravação dos logs.
    """
    # Se a pasta de logs não existir, cria
    try:
        makedirs("logs", exist_ok=True)
    except PermissionError:
        msg = "Erro ao criar a pasta 'logs'. Permissão de escrita negada!"
        raise PermissionError(msg) from None

    logger_name = filename.split(".")[0]
    log_file_path = str(Path("logs") / filename)

    # Configuração de parâmetros para geração de logs
    try:
        logging.basicConfig(level=logging.CRITICAL)
        logger = logging.getLogger(logger_name)
        logger.setLevel(logging.INFO)
        filehandler = logging.FileHandler(log_file_path, mode='a')
        filehandler.setLevel(logging.INFO)
        formatter = logging.Formatter('%(asctime)s - %(levelname)s | %(funcName)s: %(message)s')
        filehandler.setFormatter(formatter)
        logger.addHandler(filehandler)
    except FileNotFoundError:
        msg = f"Não foi possível encontrar/criar o arquivo de log no caminho '{log_file_path}'. Programa abortado!"
        raise FileNotFoundError(msg) from None
    except PermissionError:
        msg = f"Não foi possível criar/acessar o arquivo de log no caminho '{log_file_path}'. Permissão de " \
              f"escrita/leitura negada. Programa abortado!"
        raise PermissionError(msg) from None

    return logger


# Para facilitar, define um logger único para todas as funções
LOGGER = make_log("LOG_MLLIB.log")


def load_env_variables(path: str = ""):
    """
    Carrega as variáveis de ambiente armazenadas no arquivo '.env'.
        :param path: Caminho para o arquivo '.env'. Caso não seja especificado, fará uma busca nas pastas filhas da
                     pasta onde o programa está localizado para encontrar o arquivo '.env'.
    """
    if type(path) is str:
        if path == "":
            dotenv_path = find_dotenv(usecwd=True)
        else:
            dotenv_path = path

        if dotenv_path != "":
            msg = f"Utilizando as configurações de ambiente obtidas através do arquivo: '{dotenv_path}'"
            LOGGER.info(msg)

        load_dotenv(dotenv_path)


def get_file_s3(file_name: str, s3_server: str, access_key: str, secret_key: str, bucket: str):
    """
    Obtém um arquivo através do protocolo s3.
        :param file_name: Nome do arquivo a ser baixado.
        :param s3_server: Servidor s3 que proverá o arquivo.
        :param access_key: Chave de acesso para logar no servidor s3.
        :param secret_key: Senha para logar no servidor s3.
        :param bucket: Nome do bucket onde o arquivo a ser baixado se encontra.
        :return: Objeto contendo o arquivo baixado.
    """
    client = minio.Minio(s3_server, access_key, secret_key)

    try:
        obj_arquivo = client.get_object(bucket_name=bucket, object_name=file_name)
    except minio.error.S3Error as e:
        msg = f"Não foi possível obter o arquivo desejado. Mensagem do servidor S3: {e}"
        LOGGER.error(msg)
        raise RuntimeError(msg) from None

    return obj_arquivo.read()


def get_file_local(file_path: str):
    """
    Obtém um arquivo através do armazenamento local.
        :param file_path: Caminho do arquivo a ser carregado.
        :return: Objeto contendo o arquivo carregado.
    """
    try:
        with open(file_path, 'rb') as arq:
            bytes_arq = arq.read()
    except FileNotFoundError:
        msg = f"Não foi possível encontrar o arquivo no caminho '{file_path}'. Programa abortado!"
        LOGGER.error(msg)
        raise FileNotFoundError(msg) from None
    except PermissionError:
        msg = f"Não foi possível ler o arquivo no caminho '{file_path}'. Permissão de leitura negada. Programa " \
              f"abortado!"
        LOGGER.error(msg)
        raise PermissionError(msg) from None

    return bytes_arq


def validate_params(received_params: list, expected_params: dict) -> tuple:
    """
    Valida os parâmetros que foram recebidos via linha de comando na execução de um programa. Utiliza somente o '='
    como separador entre o parâmetro e seu valor.
        :param received_params: Parâmetros recebidos via linha de comando no momento da execução de um programa.
                                Obs.: A lista de parâmetros deve ser obtida através da propriedade 'sys.argv'.
        :param expected_params: Nomes de parâmetros e tipos esperados pelo programa que servirão como base para a
                                validação. Ex.: {'nome': str, 'idade': int}.
        :return: (True, dict). Verdadeiro, se passou na validação, e um dicionário com os nomes dos parâmetros esperados
                 e seus respectivos valores ou (False, str). Falso, se não passou na validação, e uma string com os
                 erros encontrados.
    """
    if type(expected_params) is dict:
        if len(expected_params) == 0:
            return False, "O dicionário (expected_params) contendo os parâmetros esperados está vazio."
    else:
        return False, "O tipo do parâmetro 'expected_params' não é dict."

    # Não conta o primeiro parâmetro porque é o nome do script que foi executado
    if len(received_params) <= 1:
        return False, "A lista de parâmetros recebidos (received_params) está vazia."

    parametros_validados = {}
    erros_encontrados = ""

    for p in received_params[1:]:
        nome_valor = p.split("=")

        if len(nome_valor) == 2:
            nome_param = nome_valor[0]
            valor_param = nome_valor[1]

            if nome_param in expected_params:
                try:
                    # Tenta converter para o tipo de parâmetro esperado
                    parametros_validados[nome_param] = expected_params[nome_param](valor_param)
                except ValueError:
                    erros_encontrados += f"O valor para o parâmetro '{nome_param}' deve ser do tipo " \
                                         f"{expected_params[nome_param]}.\n"
                except TypeError:
                    erros_encontrados += f"O tipo do parâmetro '{nome_param}' informado no dicionário " \
                                         f"'expected_params' está incorreto. Não digite o tipo entre aspas."
            else:
                erros_encontrados += f"O parâmetro '{nome_param}' não existe.\n"
        else:
            erros_encontrados += f"Não foi possível encontrar o separador de valores '=' no parâmetro '{p}' " \
                                 f"ou existem espaços ' = ' entre o nome do parâmetro, separador e valor.\n"

    if erros_encontrados == "":
        return True, parametros_validados
    else:
        return False, erros_encontrados


def get_models_params(path: str = "") -> dict:
    """
    Obtém os parâmetros que serão utilizados para instanciar os modelos. Será buscado um arquivo com o nome
    'params.conf' contendo o nome dos modelos como uma seção [MODEL_NAME] e os parâmetros: 'source_file',
    'model_class', 'experiment_name', 'model_provider_name' e 'dataset_provider_name'.
        :param path: Caminho onde se encontra o arquivo de parâmetros. O padrão é estar na pasta local.
        :return: Dicionário contendo como chave o nome do modelo e como valor outro dicionário com os parâmetros.
    """
    parametros_padroes = ["source_file", "model_class", "experiment_name", "model_provider_name",
                          "dataset_provider_name"]
    parametros_faltantes_por_secao = {}
    faltou_parametro = False
    conf = configparser.ConfigParser()
    param_file_path = str(Path(path) / "params.conf")

    # Carrega os parâmetros
    try:
        nome_arq_params = conf.read(param_file_path)
    except configparser.MissingSectionHeaderError as e:
        msg = f"O arquivo com os parâmetros dos modelos ('params.conf') possui seções inválidas. Programa abortado! " \
              f"Mensagem configParser: '{e}'."
        LOGGER.error(msg)
        raise ValueError(msg) from None
    except configparser.DuplicateSectionError as e:
        msg = f"Existem seções duplicadas. Programa abortado! Mensagem configParser: '{e}'."
        LOGGER.error(msg)
        raise ValueError(msg) from None
    except configparser.DuplicateOptionError as e:
        msg = f"Existem parâmetros duplicados. Programa abortado! Mensagem configParser: '{e}'."
        LOGGER.error(msg)
        raise ValueError(msg) from None

    if "params.conf" not in nome_arq_params:
        msg = "Não foi possível encontrar o arquivo com os parâmetros dos modelos ('params.conf') ou não possui " \
              "permissão para leitura. Programa abortado!"
        LOGGER.error(msg)
        raise RuntimeError(msg)

    secoes = conf.sections()

    if not secoes:
        msg = "O arquivo com os parâmetros dos modelos ('params.conf') está vazio. Programa abortado!"
        LOGGER.error(msg)
        raise ValueError(msg)

    # Verifica se tem parâmetros padrões faltantes
    for s in secoes:
        parametros_faltantes = []

        for p in parametros_padroes:
            if p not in conf[s]:
                parametros_faltantes.append(p)

        if parametros_faltantes:
            faltou_parametro = True
            parametros_faltantes_por_secao[s] = parametros_faltantes

    if faltou_parametro:
        msg = f"Um ou mais parâmetros não foram encontrados no arquivo 'params.conf'. Parâmetros faltantes por " \
              f"modelo: {parametros_faltantes_por_secao}. Programa abortado!"
        LOGGER.error(msg)
        raise ValueError(msg)

    parametros_por_modelo = {}

    # Obtém os parâmetros e gera o dicionário com os modelos e parâmetros
    for s in secoes:
        parametros_valor = {}

        for p in parametros_padroes:
            parametros_valor[p] = conf[s][p]

        parametros_por_modelo[s] = parametros_valor

    return parametros_por_modelo



# ----------------------------------------------------------------------------------------------------
# Implementação de classes que são compartilhadas entre as interfaces padrões para publicação de
# modelos no ML Worker.
#
# Uso: Implemente aqui qualquer classe que puder ser compartilhada entre as interfaces contidas no
# arquivo 'mllibprodest/interfaces.py'
# ----------------------------------------------------------------------------------------------------
import pickle
from logging import Logger
from typing import Union
from pathlib import Path
from .provider import Provider
from .utils import make_log

# Para facilitar, define um logger único para todas as funções
LOGGER = make_log("LOG_MLLIB.log")


class CommonMethods:
    """
    Métodos comuns entre as interfaces dos modelos publicados.
    """
    @staticmethod
    def make_log(filename: str) -> Logger:
        """
        Cria um arquivo para geração de logs. Se o mesmo já existir, inicia a gravação a partir do final desse arquivo.
        Também retorna um logger para gravação dos logs.
            :param filename: Nome do arquivo de logs.
            :return: Um logger para gravação dos logs.
        """
        return make_log(filename)

    @staticmethod
    def load_datasets(datasets_filenames: dict, provider: str = 'minio') -> dict:
        """
        Carrega os datasets necessários para o modelo. Os parâmetros de acesso deverão ser fornecidos através de um
        arquivo chamado '.env' que deve ser criado no repositório local e preenchido com as seguintes variáveis:
        MINIO = "nome do servidor s3", ACCESS_KEY = "chave de acesso", SECRET_KEY = "senha de acesso" e
        BUCKET = "nome do bucket", no caso do provider 'minio', ou somente a variável LOCAL_PATH = "caminho local onde
        os datasets se encontram", se o provider for 'local'. Dica de segurança: Não deixe o arquivo '.env' ser
        versionado/persistido no repositório remoto do código.
            :param datasets_filenames: Dicionário contendo os tipos de datasets e os nomes dos respectivos arquivos.
                                       Exemplo: {'features': 'nome_arquivo_features', 'targets': 'nome_arquivo_targets'}
            :param provider: Nome do provedor que fornecerá os datasets. Tipos de provider: 'minio' e 'local'.
            :return: Dicionário com os datasets carregados e prontos para serem lidos, por exemplo, através do pandas
                     com a função read_csv(), se for um arquivo csv. Obs.: As chaves do dicionário retornado serão as
                     mesmas informadas no parâmetro 'datasets_filenames' e os valores serão os datasets carregados.
        """
        return Provider.load_datasets(datasets_filenames=datasets_filenames, provider=provider)

    @staticmethod
    def load_production_params(model_name: str, provider: str = 'mlflow') -> dict:
        """
        Carrega os parâmetros que foram utilizados para treinar o modelo que está em produção.
            :param model_name: Nome do modelo que está em produção.
            :param provider: Nome do provedor que fornecerá os parâmetros do modelo que está em produção. Tipos de
                             provider: 'mlflow'.
            :return: Dicionário contendo os parâmetros carregados.
        """
        return Provider.load_production_params(model_name=model_name, provider=provider)

    @staticmethod
    def load_production_datasets_names(model_name: str, provider: str = 'mlflow') -> dict:
        """
        Carrega os nomes dos datasets que foram utilizados para treinar o modelo que está em produção.
            :param model_name: Nome do modelo que está em produção.
            :param provider: Nome do provedor que fornecerá os nomes dos datasets do modelo que está em produção.
                             Tipos de provider: 'mlflow'.
            :return: Dicionário contendo os nomes dos datasets carregados.
        """
        return Provider.load_production_datasets_names(model_name=model_name, provider=provider)

    @staticmethod
    def load_production_baseline(model_name: str, provider: str = 'mlflow') -> dict:
        """
        Carrega as métricas do modelo que está em produção que serão utilizadas como baseline para avaliação
        automatizada do modelo.
            :param model_name: Nome do modelo que está em produção.
            :param provider: Nome do provedor que fornecerá o baseline do modelo em produção. Tipos de provider:
                             'mlflow'.
            :return: Dicionário contendo as métricas de baseline.
        """
        return Provider.load_production_baseline(model_name=model_name, provider=provider)

    @staticmethod
    def load_model(model_name: str, provider: str = 'mlflow', artifacts_destination_path: str = 'temp_area'):
        """
        Carrega o modelo que está em produção e baixa os artefatos necessários.
            :param model_name: Nome do modelo que será carregado.
            :param provider: Nome do provedor que fornecerá o modelo. Tipos de provider: 'mlflow'.
            :param artifacts_destination_path: Caminho para onde os artefatos serão baixados.
            :return: Modelo carregado.
        """
        return Provider.load_model(model_name=model_name, provider=provider,
                                   artifacts_destination_path=artifacts_destination_path)

    @staticmethod
    def convert_artifact_to_pickle(model_name: str, artifact: object, file_name: str, path: str = "temp_area"):
        """
        Converte um artefato para o formato pickle, para facilitar a persistência.
            :param model_name: Nome do modelo em que o artefato será utilizado.
            :param artifact: Artefato que será convertido (str, list, dict, tuple, etc.).
            :param file_name: Nome do arquivo que será gerado (Dica: Use a extensão '.pkl').
            :param path: Caminho para gerar o artefato convertido.
        """
        caminho_artefato = str(Path(path) / model_name / file_name)

        try:
            arq = open(caminho_artefato, 'wb')
        except FileNotFoundError:
            msg = f"Não foi possível gerar o artefato '{file_name}' convertido. O caminho '{caminho_artefato}' está " \
                  f"incorreto. Programa abortado!"
            LOGGER.error(msg)
            raise FileNotFoundError(msg) from None
        except PermissionError:
            msg = f"Não foi possível gerar o artefato '{file_name}' convertido no caminho '{caminho_artefato}'. " \
                  f"Permissão de escrita negada. Programa abortado!"
            LOGGER.error(msg)
            raise PermissionError(msg) from None

        try:
            pickle.dump(artifact, arq)
        except TypeError as e:
            msg = f"Não foi possível gerar o artefato '{file_name}' com o Pickle (mensagem Pickle: {e}). Programa " \
                  f"abortado!"
            LOGGER.error(msg)
            raise TypeError(msg) from None

        arq.close()

    @staticmethod
    def convert_artifact_to_object(model_name: str, file_name: str, path: str = "temp_area") -> \
            Union[list, tuple, dict, object]:
        """
        Converte um artefato que está no formato pickle para o objeto de origem.
            :param model_name: Nome do modelo ao qual o artefato pertence.
            :param file_name: Nome do arquivo que será lido e convertido.
            :param path: Caminho onde o arquivo a ser convertido se encontra.
            :return: Artefato convertido.
        """
        caminho_artefato = str(Path(path) / model_name / file_name)

        try:
            arq = open(caminho_artefato, 'rb')
        except FileNotFoundError:
            msg = f"Não foi possível converter o artefato '{file_name}'. O caminho '{caminho_artefato}' não foi " \
                  f"encontrado. Programa abortado!"
            LOGGER.error(msg)
            raise FileNotFoundError(msg) from None
        except PermissionError:
            msg = f"Não foi possível converter o artefato '{file_name}' usando o caminho '{caminho_artefato}'. " \
                  f"Permissão de leitura negada. Programa abortado!"
            LOGGER.error(msg)
            raise PermissionError(msg) from None

        try:
            objeto = pickle.load(arq)
        except pickle.UnpicklingError as e:
            msg = f"Não foi possível converter o artefato '{file_name}' com o Pickle (mensagem Pickle: {e}). " \
                  f"Programa abortado!"
            LOGGER.error(msg)
            raise RuntimeError(msg)

        arq.close()

        return objeto



# ----------------------------------------------------------------------------------------------------
# Provider para obtenção de modelos e artefatos registrados no Mlflow
# ----------------------------------------------------------------------------------------------------
import mlflow
import pickle
from pathlib import Path
from mlflow.exceptions import RestException, MlflowException
from shutil import rmtree
from ..utils import make_log

# Para facilitar, define um logger único para todas as funções
LOGGER = make_log("LOG_MLLIB.log")


def load_model_mlflow(model_name: str, artifacts_destination_path: str = "temp_area"):
    """
    Carrega o modelo que está em produção e baixa os artefatos necessários utilizando o MLflow.
        :param model_name: Nome do modelo que será carregado.
        :param artifacts_destination_path: Caminho local para onde os artefatos serão baixados.
        :return: Modelo carregado.
    """
    artefatos_obrigatorios = ["TrainingParams.pkl", "TrainingDatasetsNames.pkl", "BaselineMetrics.pkl"]
    caminho_artefatos = Path(artifacts_destination_path) / model_name

    # Antes de carregar o modelo, faz uma limpeza em alguns artefatos antigos. Obs.: Isso é necessário para evitar que,
    # caso o usuário esqueça de salvar um artefato obrigatório, seja carregado um artefato antigo salvo localmente.
    if artifacts_destination_path == "temp_area":
        rmtree(caminho_artefatos, ignore_errors=True)
    else:
        for artefato in artefatos_obrigatorios:
            Path.unlink(caminho_artefatos / artefato, missing_ok=True)

    try:
        Path.mkdir(caminho_artefatos, parents=True, exist_ok=True)
    except PermissionError:
        msg = f"Não foi possível criar a pasta de destino dos artefatos '{artifacts_destination_path}'. Permissão " \
              f"de escrita negada. Programa abortado!"
        LOGGER.error(msg)
        raise PermissionError(msg) from None

    stage = 'Production'

    # Carrega o modelo que está em produção
    try:
        modelo = mlflow.pyfunc.load_model(model_uri=f"models:/{model_name}/{stage}")
    except RestException:
        msg = f"O modelo '{model_name}' no estágio '{stage}' não foi encontrado. Programa abortado!"
        LOGGER.error(msg)
        raise RuntimeError(msg) from None
    except MlflowException as e:
        msg = f"Não foi possível carregar o modelo '{model_name}'. Mensagem do MLFlow: '{e}'. Programa abortado!"
        LOGGER.error(msg)
        raise RuntimeError(msg) from None

    # Baixa todos os artefatos com base no 'run_id' do modelo
    endereco_base_artefatos = f"runs:/{modelo.metadata.run_id}/"

    try:
        mlflow.artifacts.download_artifacts(artifact_uri=endereco_base_artefatos, dst_path=str(caminho_artefatos))
    except MlflowException as e:
        msg = f"Não foi possível carregar os artefatos no endereço '{endereco_base_artefatos}'. " \
              f"Mensagem do MLFlow: '{e}'. Programa abortado!"
        LOGGER.error(msg)
        raise RuntimeError(msg) from None

    return modelo


def load_production_params_mlflow(model_name: str) -> dict:
    """
    Carrega os parâmetros que foram utilizados para treinar o modelo que está em produção no MLflow.
        :param model_name: Nome do modelo que está em produção.
        :return: Dicionário contendo os parâmetros carregados.
    """
    modelo = load_model_mlflow(model_name, artifacts_destination_path='temp_area')
    LOGGER.info(f"Utilizando os parâmetros de produção do modelo '{model_name}' (run_id: {modelo.metadata.run_id})")
    parametros = None
    nome_arq = str(Path("temp_area") / model_name / "TrainingParams.pkl")
    arq = None
    msg = ""

    try:
        arq = open(nome_arq, 'rb')
    except FileNotFoundError:
        msg += f"O arquivo '{nome_arq}' não foi encontrado. "
    except PermissionError:
        msg += f"Permissão de leitura negada para o arquivo '{nome_arq}'. "

    if arq is not None:
        try:
            parametros = pickle.load(arq)
        except pickle.UnpicklingError as e:
            msg += f"Não foi possível carregar os parâmetros através do arquivo '{nome_arq}' com o Pickle (mensagem " \
                   f"Pickle: {e}). "

        arq.close()

    if parametros is not None and type(parametros) is dict:
        return parametros
    else:
        msg += f"Não foi possível carregar os parâmetros de produção do modelo '{model_name}'. Certifique-se que o " \
               f"modelo exista e que possua os parâmetros persistidos num dicionário, através do Pickle, com o nome " \
               f"'TrainingParams.pkl'. Programa abortado!"
        LOGGER.error(msg)
        raise RuntimeError(msg)


def load_production_datasets_names_mlflow(model_name: str) -> dict:
    """
    Carrega os nomes dos datasets que foram utilizados para treinar o modelo que está em produção no MLflow.
        :param model_name: Nome do modelo que está em produção.
        :return: Dicionário contendo os nomes dos datasets carregados.
    """
    modelo = load_model_mlflow(model_name, artifacts_destination_path='temp_area')
    LOGGER.info(f"Utilizando os nomes dos datasets de produção do modelo '{model_name}' (run_id: "
                f"{modelo.metadata.run_id})")
    nomes_datasets = None
    nome_arq = str(Path("temp_area") / model_name / "TrainingDatasetsNames.pkl")
    arq = None
    msg = ""

    try:
        arq = open(nome_arq, 'rb')
    except FileNotFoundError:
        msg += f"O arquivo '{nome_arq}' não foi encontrado. "
    except PermissionError:
        msg += f"Permissão de leitura negada para o arquivo '{nome_arq}'. "

    if arq is not None:
        try:
            nomes_datasets = pickle.load(arq)
        except pickle.UnpicklingError as e:
            msg += f"Não foi possível carregar os nomes dos datasets através do arquivo '{nome_arq}' com o Pickle " \
                   f"(mensagem Pickle: {e}). "

        arq.close()

    if nomes_datasets is not None and type(nomes_datasets) is dict:
        return nomes_datasets
    else:
        msg += f"Não foi possível carregar os nomes dos datasets de produção do modelo '{model_name}'. Certifique-se " \
               f"que o modelo exista e que possua os nomes dos datasets persistidos num dicionário, através do " \
               f"Pickle, com o nome 'TrainingDatasetsNames.pkl'. Programa abortado!"
        LOGGER.error(msg)
        raise RuntimeError(msg)


def load_production_baseline_mlflow(model_name: str) -> dict:
    """
    Carrega as métricas do modelo que está em produção no MLflow que serão utilizadas como baseline para avaliação
    automatizada do modelo.
        :param model_name: Nome do modelo que está em produção.
        :return: Dicionário contendo as métricas de baseline.
    """
    modelo = load_model_mlflow(model_name, artifacts_destination_path='temp_area')
    LOGGER.info(f"Utilizando o baseline de produção do modelo '{model_name}' (run_id: {modelo.metadata.run_id})")
    baseline = None
    nome_arq = str(Path("temp_area") / model_name / "BaselineMetrics.pkl")
    arq = None
    msg = ""

    try:
        arq = open(nome_arq, 'rb')
    except FileNotFoundError:
        msg += f"O arquivo '{nome_arq}' não foi encontrado. "
    except PermissionError:
        msg += f"Permissão de leitura negada para o arquivo '{nome_arq}'. "

    if arq is not None:
        try:
            baseline = pickle.load(arq)
        except pickle.UnpicklingError as e:
            msg += f"Não foi possível carregar o baseline através do arquivo '{nome_arq}' com o Pickle (mensagem " \
                   f"Pickle: {e}). "

        arq.close()

    if baseline is not None and type(baseline) is dict:
        return baseline
    else:
        msg += f"Não foi possível carregar o baseline de produção do modelo '{model_name}'. Certifique-se que o " \
               f"modelo exista e que possua o baseline persistido num dicionário, através do Pickle, com o nome " \
               f"'BaselineMetrics.pkl'. Programa abortado!"
        LOGGER.error(msg)
        raise RuntimeError(msg)



# ---------------------------------------------------------------------------------------------------------
# Classes e funções para realização de testes.
#
# Obs.: Não tem nada a ver com 'Unit testing framework' (https://docs.python.org/3/library/unittest.html)
# ---------------------------------------------------------------------------------------------------------
import hashlib
import time
import os
from pathlib import Path
from shutil import copytree, rmtree
from ..utils import make_log
from ..initiators.model_initiator import InitModels

# Códigos para impressão de mensagens coloridas no terminal
BLUE = "\033[1;34m"
RED = "\033[1;31m"
GREEN = "\033[0;32m"
BOLD = "\033[;1m"
RESET = "\033[0;0m"


class Test:
    """
    Classe para realização de testes para validação da implementação das interfaces da lib. Obs.: Não tem nada a ver
    com 'Unit testing framework' (https://docs.python.org/3/library/unittest.html).
    """
    def __init__(self):
        self.__logger = make_log("LOG_TESTS.log")

        h = hashlib.md5()
        h.update(str(time.time()).encode())
        self.__test_id = h.hexdigest()

        msg = f"------------------> Instanciando o teste com o ID: {self.__test_id} <------------------"
        self.__logger.info(msg)
        print(f"\n\n      {BOLD}{BLUE}{msg}{RESET}")

        pasta_corrente = Path.cwd().name

        if pasta_corrente == "worker_pub":
            self.__nome_pasta_worker = "worker_pub"
            self.__nome_mytest = "mytest_pub.py"
        elif pasta_corrente == "worker_retrain":
            self.__nome_pasta_worker = "worker_retrain"
            self.__nome_mytest = "mytest_retrain.py"
        else:
            msg = "Não foi possível encontrar a pasta 'worker_pub' ou 'worker_retrain'. Verifique se estas pastas " \
                  "existem ou se você está rodando os testes na pasta correta. Teste abortado!"
            self.__logger.error(msg)
            print(f"\n\n{msg}")
            print(f"\n{RED}  *** O TESTE FALHOU! ***{RESET}\n")
            exit(1)

        if Path.exists(Path("requirements.txt")):
            arq = None

            try:
                arq = open("requirements.txt", "r")
            except PermissionError:
                msg = f"Não foi possível ler o arquivo 'requirements.txt' dentro da pasta " \
                      f"'{self.__nome_pasta_worker}'. Permissão de leitura negada. Teste abortado!"
                self.__logger.error(msg)
                print(f"\n\n{msg}")
                print(f"\n{RED}  *** O TESTE FALHOU! ***{RESET}\n")
                exit(1)

            reqs = arq.readlines()
            arq.close()
            vazio = True

            for r in reqs:
                linha = r.rstrip('\n')

                if len(linha) >= 1:
                    vazio = False
                    break

            if not vazio:
                msg = f"O arquivo 'requirements.txt' foi encontrado na pasta '{self.__nome_pasta_worker}' e NÃO " \
                      f"está vazio!"
                self.__logger.info(msg)
                print(f"\n\n{msg}")
            else:
                msg = f"O arquivo 'requirements.txt' foi encontrado na pasta '{self.__nome_pasta_worker}' mas está " \
                      f"vazio. Teste abortado!"
                self.__logger.error(msg)
                print(f"\n\n{msg}")
                print(f"\n{RED}  *** O TESTE FALHOU! ***{RESET}\n")
                exit(1)
        else:
            msg = f"O arquivo 'requirements.txt' não foi encontrado na pasta '{self.__nome_pasta_worker}'. Por favor," \
                  f" gere este arquivo com a lista de pacotes que foram utilizados na construção do(s) modelo(s). " \
                  f"Teste abortado!"
            self.__logger.error(msg)
            print(f"\n\n{msg}")
            print(f"\n{RED}  *** O TESTE FALHOU! ***{RESET}\n")
            exit(1)

        # Howto para informar como passar uma função personalizada para realização dos testes
        self.__howto_msg = f"1. Edite o script '{self.__nome_mytest}' (está na pasta '{self.__nome_pasta_worker}')" \
                           f" para implementar a função 'test()';\n2. Crie os testes adicionais que você julgar " \
                           f"necessários para testar o código do worker.\n   REGRAS: Esta função não deve receber " \
                           f"parâmetros adicionais. Não pode retornar valores. Todos os dados para rodá-la devem ser " \
                           f"obtidos de dentro dela.\n   A responsabilidade de criar os testes personalizados e " \
                           f"verificar se passaram é do desenvolvedor do modelo."

        # Artefatos obrigatórios que devem ter os tipos de valores de retorno testados
        self.__mandatory_artifacts = {
            'TrainingParams.pkl': dict,
            'TrainingDatasetsNames.pkl': dict,
            'BaselineMetrics.pkl': dict
        }

        # Métodos que devem ter os tipos de valores de retorno testados
        self.__methods_to_test = {
            'ModeloCLF': {
                'ModeloCLF.get_model_name': str,
                'ModeloCLF.get_model_provider_name': str,
                'ModeloCLF.get_model_info': dict
            },
            'ModeloRETRAIN': {
                'ModeloRETRAIN.get_model_name': str,
                'ModeloRETRAIN.get_model_provider_name': str,
                'ModeloRETRAIN.get_experiment_name': str,
                'ModeloRETRAIN.get_dataset_provider_name': str
            }
        }

        # Guarda os modelos instanciados
        self.__modelos = None

        # Instancia os scripts de testes personalizados
        self.__validation_function = None

        if Path.exists(Path(self.__nome_mytest)):
            try:
                if self.__nome_mytest == "mytest_pub.py":
                    from mytest_pub import test
                    self.__validation_function = test
                elif self.__nome_mytest == "mytest_retrain.py":
                    from mytest_retrain import test
                    self.__validation_function = test
            except ImportError:
                msg = "AVISO: A função para realização do teste personalizado pelo usuário não foi encontrada." \
                      "\n       Caso deseje criar um teste personalizado, faça o seguinte:\n"
                self.__logger.info(msg + self.__howto_msg)
                print(f"\n\n{msg + self.__howto_msg}\n")
        else:
            msg = f"AVISO: O arquivo '{self.__nome_mytest}' não foi encontrado na pasta '{self.__nome_pasta_worker}'" \
                  f". Caso deseje criar um teste personalizado, crie\n       um arquivo chamado " \
                  f"'{self.__nome_mytest}' e faça o seguinte:\n"
            self.__logger.info(msg + self.__howto_msg)
            print(f"\n\n{msg + self.__howto_msg}\n")

    def __validate_methods(self, model, class_name: str) -> bool:
        """
        Valida os métodos definidos no construtor da classe através do atributo 'self.__methods_to_test'.
            :param model: Modelo de onde os métodos serão chamados.
            :param class_name: Nome da classe que possui os métodos que serão validados.
            :return: True, se todos os métodos foram validados. False, caso algum método não tenha sido validado.
        """
        validado = True

        for method_name, return_type in self.__methods_to_test[class_name].items():
            nome_metodo_aux = method_name.split(".")[1]
            msg = f"=> Testando o método '{nome_metodo_aux}'..."
            self.__logger.info(msg)
            print(f"\n\n{msg}\n")

            retorno = None

            try:
                if method_name == "ModeloCLF.get_model_name":
                    retorno = model.get_model_name()
                elif method_name == "ModeloCLF.get_model_provider_name":
                    retorno = model.get_model_provider_name()
                elif method_name == "ModeloCLF.get_model_info":
                    retorno = model.get_model_info()
                elif method_name == "ModeloRETRAIN.get_model_name":
                    retorno = model.get_model_name()
                elif method_name == "ModeloRETRAIN.get_model_provider_name":
                    retorno = model.get_model_provider_name()
                elif method_name == "ModeloRETRAIN.get_experiment_name":
                    retorno = model.get_experiment_name()
                elif method_name == "ModeloRETRAIN.get_dataset_provider_name":
                    retorno = model.get_dataset_provider_name()
            except AttributeError as e:
                msg = f"A chamada ao método '{nome_metodo_aux}' falhou: {str(e)}."
                self.__logger.error(msg)
                print(f"\n{RED}***ERRO***: {msg}{RESET}\n")
                print(f"\n{RED}  *** O TESTE FALHOU! ***{RESET}\n")
                exit(1)

            tipo_retorno = type(retorno)

            if tipo_retorno is return_type:
                msg = f"Tipo de retorno da chamada ao método '{nome_metodo_aux}' OK: {retorno}"
                self.__logger.info(msg)
                print(f"   {msg}\n")
            else:
                validado = False
                msg = f"O método '{nome_metodo_aux}' deve retornar o tipo '{return_type.__name__}', porém retornou " \
                      f"'{tipo_retorno.__name__}'."
                self.__logger.error(msg)
                print(f"{RED}***ERRO***: {msg}{RESET}\n")

        return validado

    def __validate_mandatory_artifacts(self, model) -> bool:
        """
        Valida se os artefatos obrigatórios definidos para o modelo foram implementados pelo usuário.
            :param model: Modelo de onde será chamada a função utilitária 'convert_artifact_to_object'.
            :return: True, se todos os artefatos obrigatórios foram encontrados. False, caso algum artefato obrigatório
                     não tenha sido encontrado.
        """
        validado = True

        for nome_artefato_obrigatorio, tipo_artefato_obrigatorio in self.__mandatory_artifacts.items():
            msg = f"=> Procurando pelo artefato obrigatório '{nome_artefato_obrigatorio}'..."
            self.__logger.info(msg)
            print(f"\n{msg}\n")

            artefato_lido = model.convert_artifact_to_object(model_name=model.get_model_name(),
                                                             file_name=nome_artefato_obrigatorio)
            tipo_artefato_lido = type(artefato_lido)

            if tipo_artefato_lido is tipo_artefato_obrigatorio:
                if len(artefato_lido) > 0:
                    msg = f"Conteúdo do artefato obrigatório '{nome_artefato_obrigatorio}': '{artefato_lido}'"
                    self.__logger.info(msg)
                    print(f"  {msg}\n")
                else:
                    validado = False
                    msg = f"O artefato obrigatório '{nome_artefato_obrigatorio}' deve ter o tamanho maior que 0 (zero)."
                    self.__logger.error(msg)
                    print(f"{RED}***ERRO***: {msg}{RESET}\n")
            else:
                validado = False
                msg = f"O artefato obrigatório '{nome_artefato_obrigatorio}' deve ser do tipo " \
                      f"'{tipo_artefato_obrigatorio.__name__}', porém é do tipo: '{tipo_artefato_lido.__name__}'."
                self.__logger.error(msg)
                print(f"{RED}***ERRO***: {msg}{RESET}\n")

        return validado

    def __validate_pub(self):
        """
        Valida os scripts para publicação do modelo utilizando a classe ModeloCLF.
        """
        msg = "Scripts para publicação do modelo utilizando a classe 'ModeloCLF'..."
        self.__logger.info(msg)
        print(f"\n\n # {msg}")

        for nome_modelo, modelo in self.__modelos.items():
            print(f"\n{GREEN} ### MODELO: {nome_modelo} ### {RESET}\n")

            if not self.__validate_methods(modelo, "ModeloCLF"):
                msg = "Erro ao validar os métodos obrigatórios. Verifique as mensagens de validação acima."
                self.__logger.error(msg)
                print(f"\n\n{RED} *** {msg}{RESET}\n")
                print(f"\n{RED}  *** O TESTE FALHOU! ***{RESET}\n")
                exit(1)

        msg = "AVISO: Os métodos 'predict' e 'evaluate' não serão testados porque necessitam de dados que " \
              "são específicos para cada implementação.\nCaso deseje testar estes métodos, faça o seguinte:\n"
        self.__logger.info(msg + self.__howto_msg)
        print(f"\n\n{msg + self.__howto_msg}\n")

    def __validate_retrain(self):
        """
        Valida os scripts para publicação do modelo utilizando a classe ModeloRETRAIN.
        """
        msg = "Scripts para publicação do modelo utilizando a classe 'ModeloRETRAIN'..."
        self.__logger.info(msg)
        print(f"\n\n # {msg}")

        for nome_modelo, modelo in self.__modelos.items():
            print(f"\n{GREEN} ### MODELO: {nome_modelo} ### {RESET}\n")

            if not self.__validate_methods(modelo, "ModeloRETRAIN"):
                msg = "Erro ao validar os métodos obrigatórios. Verifique as mensagens de validação acima."
                self.__logger.error(msg)
                print(f"\n\n{RED} *** {msg}{RESET}\n")
                print(f"\n{RED}  *** O TESTE FALHOU! ***{RESET}\n")
                exit(1)

            # Testa o carregamento do modelo e verifica se os artefatos obrigatórios estão de acordo
            model_name = modelo.get_model_name()
            provider_name = modelo.get_model_provider_name()

            try:
                modelo.load_model(model_name=model_name, provider=provider_name)
            except AttributeError as e:
                msg = f"A chamada ao método 'load_model' falhou: {str(e)}."
                self.__logger.error(msg)
                print(f"\n\n{msg}\n")
                print(f"\n{RED}  *** O TESTE FALHOU! ***{RESET}\n")
                exit(1)

            if not self.__validate_mandatory_artifacts(modelo):
                msg = "Erro ao validar os artefatos obrigatórios. Verifique as mensagens de validação acima."
                self.__logger.error(msg)
                print(f"\n\n{RED} *** {msg}{RESET}\n")
                print(f"\n{RED}  *** O TESTE FALHOU! ***{RESET}\n")
                exit(1)

        msg = "AVISO: Os métodos 'evaluate' e 'retrain' não serão testados porque necessitam de dados que " \
              "são específicos para cada implementação.\nCaso deseje testar estes métodos, faça o seguinte:\n"
        self.__logger.info(msg + self.__howto_msg)
        print(f"\n\n{msg + self.__howto_msg}\n")

    def __validate_params_returns(self):
        """
        Valida alguns parâmetros e valores de retorno solicitados pelas interfaces que foram implementados. Porém, não
        é escopo desta função validar os dados que estão sendo utilizados pelas funções implementadas.
        """
        msg = f"=> INICIO: Test ID={self.__test_id}. Teste padrão da lib."
        self.__logger.info(msg)
        print(f"\n\n{BOLD}{BLUE}{msg}{RESET}\n")

        if self.__nome_pasta_worker == "worker_pub":
            self.__validate_pub()
        elif self.__nome_pasta_worker == "worker_retrain":
            self.__validate_retrain()

        msg = f"=> FIM: Test ID={self.__test_id}. Teste padrão da lib."
        self.__logger.info(msg)
        print(f"\n\n{BOLD}{BLUE}{msg}{RESET}\n")

    def __run_validation_function(self):
        """
        Executa a função de validação criada e personalizada pelo usuário.
        """
        msg = f"=> INICIO: Test ID={self.__test_id}. Testes personalizados pelo usuário. Dados da função utilizada: " \
              f"{str(self.__validation_function)}."
        self.__logger.info(msg)
        print(f"\n\n{BOLD}{BLUE}{msg}{RESET}\n")

        type_validation_function = type(self.__validation_function).__name__

        if type_validation_function != 'function':
            msg = f"O parâmetro 'validation_function' deve receber uma função mas recebeu " \
                  f"'{type_validation_function}'. Verifique a implementação da função 'test' no arquivo " \
                  f"'{self.__nome_mytest}'."
            self.__logger.error(msg)
            print(f"\n\n{msg}")
            print(f"\n{RED}  *** O TESTE FALHOU! ***{RESET}\n")
            exit(1)

        self.__validation_function(self.__modelos)

        msg = f"=> FIM: Test ID={self.__test_id}. Testes personalizados pelo usuário."
        self.__logger.info(msg)
        print(f"\n\n{BOLD}{BLUE}{msg}{RESET}\n")

    def validate(self, mlruns_path: str = ""):
        """
        Valida a implementação das interfaces da lib através das chamadas de funções específicas para este fim. Além do
        conjunto de testes padrões executados pela lib, se os arquivos 'mytest_pub.py' e 'mytest_retrain.py' existirem
        nas pastas de cada worker respectivamente, a função 'test()' será procurada e, caso exista, será carregada e
        utilizada para rodar os testes personalizados pelo usuário.
            :param mlruns_path: Caminho completo da pasta 'mlruns'. Esta pasta é criada na pasta local onde o servidor
                                do MLflow, utilizado para registrar o modelo, foi iniciado. Se o valor padrão for
                                mantido, tenta utilizar as variáveis de ambiente 'MLFLOW_TRACKING_URI', para encontrar
                                o endereço do servidor MLflow, e 'MLFLOW_S3_ENDPOINT_URL' para obter o endereço do
                                endpoint que é utilizado para a persistência dos artefatos gerados nos experimentos.
        """
        # Verifica se as variáveis de ambiente para o MLflow funcionar remotamente e com os artefatos via s3 existem.
        # Se não existirem, prepara o ambiente para rodar os testes com o MLflow local.
        if os.getenv("MLFLOW_TRACKING_URI") is None and os.getenv("MLFLOW_S3_ENDPOINT_URL") is None:
            if mlruns_path != "":
                if "mlruns" in str(Path(mlruns_path)):
                    caminho_mlruns = str(Path(mlruns_path))
                else:
                    caminho_mlruns = str(Path(mlruns_path) / "mlruns")

                try:
                    copytree(caminho_mlruns, "mlruns", dirs_exist_ok=True)
                except FileNotFoundError:
                    msg = "O caminho para a pasta 'mlruns' está incorreto. Teste abortado!"
                    self.__logger.error(msg)
                    print(f"\n\n{msg}")
                    print(f"\n{RED}  *** O TESTE FALHOU! ***{RESET}\n")
                    exit(1)
                except PermissionError:
                    msg = f"Permissão de leitura na pasta de origem ('{caminho_mlruns}') ou gravação na pasta de " \
                          f"destino ('mlruns') negada. Teste abortado!"
                    self.__logger.error(msg)
                    print(f"\n\n{msg}")
                    print(f"\n{RED}  *** O TESTE FALHOU! ***{RESET}\n")
                    exit(1)

                os.environ["MLFLOW_TRACKING_URI"] = "http://localhost:5000"

                msg = "=> Instanciando os modelos definidos no arquivo 'params.conf'..."
                self.__logger.info(msg)
                print(f"\n\n{msg}\n")

                # Não tratei as exceções para preservar o traceback e facilitar a procura da origem do erro.
                self.__modelos = InitModels.init_models()

                # Roda os testes padrões da lib
                self.__validate_params_returns()

                # Se foi criada uma função para o teste, utiliza
                if self.__validation_function is not None:
                    self.__run_validation_function()

                # Atualiza a pasta 'mlruns'
                try:
                    copytree("mlruns", caminho_mlruns, dirs_exist_ok=True)
                except PermissionError:
                    msg = f"Permissão de leitura na pasta de origem ('mlruns') ou gravação na pasta de destino " \
                          f"('{caminho_mlruns}') negada. Teste abortado!"
                    self.__logger.error(msg)
                    print(f"\n\n{msg}")
                    print(f"\n{RED}  *** O TESTE FALHOU! ***{RESET}\n")
                    exit(1)

                # Limpa os arquivos criados na execução dos testes
                for path in Path("temp_area").glob("**/*"):
                    if path.is_file():
                        path.unlink()
                    elif path.is_dir():
                        rmtree(path)

                rmtree("mlruns", ignore_errors=True)
            else:
                msg = "Informe o parâmetro '--mlruns_path=<caminho completo para a pasta mlruns>' ao rodar o script. " \
                      "A pasta 'mlruns' é criada na pasta local onde o servidor do MLflow, utilizado para registrar " \
                      "o modelo, foi iniciado. Teste abortado!"
                self.__logger.error(msg)
                print(f"\n\n{msg}")
                print(f"\n{RED}  *** O TESTE FALHOU! ***{RESET}\n")
                exit(1)
        else:  # Encontrou as variáveis de ambiente e vai rodar com o MLflow remoto
            msg = "=> Instanciando os modelos definidos no arquivo 'params.conf'..."
            self.__logger.info(msg)
            print(f"\n\n{msg}\n")

            # Não tratei as exceções para preservar o traceback e facilitar a procura da origem do erro.
            self.__modelos = InitModels.init_models()

            self.__validate_params_returns()

            if self.__validation_function is not None:
                self.__run_validation_function()



# ----------------------------------------------------------------------------------------------------
# Funções para prover o acesso aos modelos persistidos e aos datasets
# ----------------------------------------------------------------------------------------------------
from .utils import make_log
from .providers_types.minio_provider import load_datasets_minio
from .providers_types.local_provider import load_datasets_local
from .providers_types.mlflow_provider import load_production_params_mlflow, load_production_datasets_names_mlflow, \
    load_production_baseline_mlflow, load_model_mlflow

# Para facilitar, define um logger único para todas as funções
LOGGER = make_log("LOG_MLLIB.log")


class Provider:
    """
    Classe para prover o acesso aos modelos persistidos e aos datasets.
    """
    @staticmethod
    def load_datasets(datasets_filenames: dict, provider: str = 'minio') -> dict:
        """
        Carrega os datasets necessários para o modelo.
            :param datasets_filenames: Dicionário contendo os tipos de datasets e os nomes dos respectivos arquivos.
                                       Exemplo: {'features': 'nome_arquivo_features', 'targets': 'nome_arquivo_targets'}
            :param provider: Nome do provedor que fornecerá os datasets. Tipos de provider: 'minio' e 'local'.
            :return: Dicionário com os datasets carregados.
        """
        if provider == "minio":
            return load_datasets_minio(datasets_filenames)
        elif provider == "local":
            return load_datasets_local(datasets_filenames)
        else:
            msg = f"Não foi possível carregar os datasets. O provider '{provider}' não foi encontrado. Programa " \
                  f"abortado!"
            LOGGER.error(msg)
            raise RuntimeError(msg)

    @staticmethod
    def load_production_params(model_name: str, provider: str = 'mlflow') -> dict:
        """
        Carrega os parâmetros que foram utilizados para treinar o modelo que está em produção.
            :param model_name: Nome do modelo que está em produção.
            :param provider: Nome do provedor que fornecerá os parâmetros do modelo em produção. Tipos de provider:
                             'mlflow'.
            :return: Dicionário contendo os parâmetros carregados.
        """
        if provider == "mlflow":
            return load_production_params_mlflow(model_name)
        else:
            msg = f"Não foi possível carregar os parâmetros do modelo '{model_name}'. O provider '{provider}' não " \
                  f"foi encontrado. Programa abortado!"
            LOGGER.error(msg)
            raise RuntimeError(msg)

    @staticmethod
    def load_production_datasets_names(model_name: str, provider: str = 'mlflow') -> dict:
        """
        Carrega os nomes dos datasets que foram utilizados para treinar o modelo que está em produção.
            :param model_name: Nome do modelo que está em produção.
            :param provider: Nome do provedor que fornecerá os nomes dos datasets do modelo em produção. Tipos de
                             provider: 'mlflow'.
            :return: Dicionário contendo os nomes dos datasets carregados.
        """
        if provider == "mlflow":
            return load_production_datasets_names_mlflow(model_name)
        else:
            msg = f"Não foi possível carregar os nomes dos datasets do modelo '{model_name}'. O provider " \
                  f"'{provider}' não foi encontrado. Programa abortado!"
            LOGGER.error(msg)
            raise RuntimeError(msg)

    @staticmethod
    def load_production_baseline(model_name: str, provider: str = 'mlflow') -> dict:
        """
        Carrega as métricas do modelo que está em produção que serão utilizadas como baseline para avaliação
        automatizada do modelo.
            :param model_name: Nome do modelo que está em produção.
            :param provider: Nome do provedor que fornecerá o baseline do modelo em produção. Tipos de provider:
                             'mlflow'.
            :return: Dicionário contendo as métricas de baseline.
        """
        if provider == "mlflow":
            return load_production_baseline_mlflow(model_name)
        else:
            msg = f"Não foi possível carregar o baseline do modelo '{model_name}'. O provider '{provider}' não " \
                  f"foi encontrado. Programa abortado!"
            LOGGER.error(msg)
            raise RuntimeError(msg)

    @staticmethod
    def load_model(model_name: str, provider: str = 'mlflow', artifacts_destination_path: str = 'temp_area'):
        """
        Carrega o modelo que está em produção e baixa os artefatos necessários.
            :param model_name: Nome do modelo que será carregado.
            :param provider: Nome do provedor que fornecerá o modelo. Tipos de provider: 'mlflow'.
            :param artifacts_destination_path: Caminho para onde os artefatos serão baixados.
            :return: Modelo carregado.
        """
        if provider == "mlflow":
            return load_model_mlflow(model_name, artifacts_destination_path)
        else:
            msg = f"Não foi possível carregar o modelo '{model_name}'. O provider '{provider}' não foi encontrado. " \
                  f"Programa abortado!"
            LOGGER.error(msg)
            raise RuntimeError(msg)



# ----------------------------------------------------------------------------------------------------
# Implementação das interfaces padrões para publicação de modelos no ML Worker.
#
# A função deste arquivo é a definição das interfaces padrões que devem ser utilizadas para fazer a
# publicação dos modelos que serão consumidos através do ML Worker.
#
# Uso: Distribua estas definições para os interessados em publicar os modelos.
#
# Fonte: https://realpython.com/python-interface/
# Baseado na seção: "Using Abstract Method Declaration"
# ----------------------------------------------------------------------------------------------------
import abc
from .shared_classes import CommonMethods
from math import ceil


class ModelPublicationInterfaceCLF(CommonMethods, metaclass=abc.ABCMeta):
    """
    Interface para publicação de modelos de ML com a função de classificação.
    Leia a documentação dos métodos e implemente-os seguindo as recomendações de tipos de parâmetros e retorno.
    """
    @classmethod
    def __subclasshook__(cls, subclass):
        return (hasattr(subclass, 'get_model_name') and
                callable(subclass.get_model_name) and
                hasattr(subclass, 'get_model_provider_name') and
                callable(subclass.get_model_provider_name) and
                hasattr(subclass, 'predict') and
                callable(subclass.predict) and
                hasattr(subclass, 'evaluate') and
                callable(subclass.evaluate) and
                hasattr(subclass, 'get_model_info') and
                callable(subclass.get_model_info) or
                NotImplemented)

    @abc.abstractmethod
    def get_model_name(self) -> str:
        """
        Obtém o nome do modelo que está em produção.
            :return: Nome do modelo que está em produção.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def get_model_provider_name(self) -> str:
        """
        Obtém o nome do provider que proverá o modelo que está em produção.
            :return: Nome do provider que proverá o modelo que está em produção.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def predict(self, dataset: list) -> list:
        """
        Faz predições utilizando o modelo que está em produção.
            :param dataset: Lista com os dados utilizados como features para realizar a predição.
            :return: Lista com os labels preditos.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def evaluate(self, data_features: list, data_targets: list) -> dict:
        """
        Calcula as métricas para avaliação do modelo.
            :param data_features: Lista com os dados utilizados como features para realizar o cálculo.
            :param data_targets: Lista com os targets correspondentes às features para realizar o cálculo.
                                 É imprescindível que a posição de cada elemento da lista de features corresponda à
                                 resposta esperada para cada elemento da lista de targets.
            :return: Dicionário com as métricas calculadas.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def get_model_info(self) -> dict:
        """
        Obtém informações sobre o modelo que está em produção.
            :return: Dicionário com informações sobre o modelo.
        """
        raise NotImplementedError


class ModelPublicationInterfaceRETRAIN(CommonMethods, metaclass=abc.ABCMeta):
    """
    Interface para publicação de modelos de ML com a função de retreino.
    Leia a documentação dos métodos e implemente-os seguindo as recomendações de tipos de parâmetros e retorno.
    """
    @classmethod
    def __subclasshook__(cls, subclass):
        return (hasattr(subclass, 'get_model_name') and
                callable(subclass.get_model_name) and
                hasattr(subclass, 'get_model_provider_name') and
                callable(subclass.get_model_provider_name) and
                hasattr(subclass, 'get_experiment_name') and
                callable(subclass.get_experiment_name) and
                hasattr(subclass, 'get_dataset_provider_name') and
                callable(subclass.get_dataset_provider_name) and
                hasattr(subclass, 'evaluate') and
                callable(subclass.evaluate) and
                hasattr(subclass, 'retrain') and
                callable(subclass.retrain) or
                NotImplemented)

    @abc.abstractmethod
    def get_model_name(self) -> str:
        """
        Obtém o nome do modelo que está em produção.
            :return: Nome do modelo que está em produção.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def get_model_provider_name(self) -> str:
        """
        Obtém o nome do provider que proverá o modelo que está em produção.
            :return: Nome do provider que proverá o modelo que está em produção.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def get_experiment_name(self) -> str:
        """
        Obtém o nome do experimento que será utilizado em produção.
            :return: Nome do experimento que será utilizado em produção.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def get_dataset_provider_name(self) -> str:
        """
        Obtém o nome do provider do dataset usado pelo modelo que está em produção.
            :return: Nome do provider do dataset usado pelo modelo que está em produção.
        """
        raise NotImplementedError

    @staticmethod
    def generate_batch_indices(dataset_size: int, batch_size: int = 100000):
        """
        Gera indices para auxiliar na predição de registros em lote. Esta função gera uma lista contendo tuplas de
        indices (inicio, fim) que pode ser percorrida em um laço para obtenção dos índices para fazer o fatiamento do
        dataset e auxiliar a predição em lote. Ex. de lista gerada: [(0, 5), (5, 10), (10, 15), ...].
            :param dataset_size: Tamanho do dataset que será predito.
            :param batch_size: Tamanho do lote.
            :return: Lista contendo tuplas de indices (inicio, fim).
        """
        if dataset_size <= 0:
            return []

        tamanho_minimo_lote = 10000
        indices = []  # Guarda as tuplas de indices que serão utilizadas para fatiar o dataset

        if batch_size >= tamanho_minimo_lote:
            ind_inicial = 0
            ind_final = batch_size if batch_size < dataset_size else dataset_size
            qtd_lotes = ceil(dataset_size / batch_size)

            for i in range(qtd_lotes):
                indices.append((ind_inicial, ind_final))
                ind_inicial = ind_final

                # Evita que o último indice seja maior que o tamanho do dataset
                if ind_inicial + batch_size < dataset_size:
                    ind_final += batch_size
                else:
                    ind_final = dataset_size
        else:
            indices.append((0, dataset_size))

        return indices

    @abc.abstractmethod
    def evaluate(self, model, datasets: dict, baseline_metrics: dict, training_params: dict,
                 artifacts_path: str = "temp_area", batch_size: int = 100000) -> (bool, dict):
        """
        Faz a avaliação do modelo que está em produção e compara com as métricas de baseline definidas para ele.
            :param model: Modelo que está em produção.
            :param datasets: Dicionário com os datasets que serão utilizados na avaliação. Dica: Colocar o tipo de
                             dataset (features, targets, etc.) como chave e o dataset em si como valor.
            :param baseline_metrics: Dicionário com as métricas do modelo em produção que servirão de baseline para a
                                     avaliação.
            :param training_params: Dicionário com os parâmetros utilizados no treinamento do modelo que está em
                                    produção.
            :param artifacts_path: Caminho local para a obtenção dos artefatos do modelo. Para facilitar, utilize o
                                   valor padrão 'temp_area'.
            :param batch_size: Tamanho do lote. Utilizado para datasets grandes, para não faltar memória ao realizar
                               as predições.
            :return: Tupla contendo: True se o modelo passou na avaliação das métricas ou False, caso contrário, e
                                     um dicionário com informações adicionais sobre a avaliação.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def retrain(self, production_model_name: str, production_params: dict, experiment_name: str, datasets: dict,
                reasons: dict):
        """
        Faz o retreinamento do modelo de forma automatizada.
            :param production_model_name: Nome do modelo que está em produção para obtenção dos parâmetros para o
                                          retreino, se necessário.
            :param production_params: Dicionário com os parâmetros que foram utilizados no treinamento do modelo que
                                      está em produção.
            :param experiment_name: Nome do experimento para persistir o modelo retreinado.
            :param datasets: Dicionário com os datasets que serão utilizados no retreino. Dica: Colocar o tipo de
                             dataset (features, targets, etc.) como chave e o dataset em si como valor.
            :param reasons: Dicionário com o(s) motivo(s) para realização do retreinamento e/ou informações adicionais.
        """
        raise NotImplementedError



# ---------------------------------------------------------------------------------------------------------
# Classes e funções para inicialização de modelos de ML (Machine Learning).
# ---------------------------------------------------------------------------------------------------------
import importlib
from ..utils import get_models_params, make_log

# Para facilitar, define um logger único para todas as funções
LOGGER = make_log("LOG_MLLIB.log")


class InitModels:
    """
    Classe utilizada para instanciar os modelos de ML (Machine Learning).
    """
    @staticmethod
    def init_models(path: str = "") -> dict:
        """
        Inicia os modelos de ML (Machine Learning) utilizando os parâmetros configurados no arquivo 'params.conf'.
            :param path: Caminho onde se encontra o arquivo de parâmetros. O padrão é estar na pasta local.
            :return: Dicionário contendo como chave os nomes dos modelos e como valor os modelos instanciados.
        """
        modelos = {}
        models_params = get_models_params(path)

        for model_name in models_params.keys():
            caminho_import = f"models.{models_params[model_name]['source_file']}"

            try:
                modulo = importlib.import_module(caminho_import, package=None)
            except ModuleNotFoundError as e:
                msg_erro = str(e)

                if caminho_import in msg_erro:
                    msg = f"Modelo: {model_name}. O módulo '{caminho_import}' não foi encontrado. Verifique no " \
                          f"arquivo 'params.conf' se o parâmetro 'source_file' foi informado corretamente e/ou se " \
                          f"este módulo está dentro da pasta 'models'. Programa abortado!"
                    LOGGER.error(msg)
                    raise ModuleNotFoundError(msg) from None
                else:
                    msg = f"Modelo: {model_name}. Erro ao importar os módulos necessários para o módulo " \
                          f"'{caminho_import}'. Mensagem do Import: {msg_erro}"
                    LOGGER.error(msg)
                    # Não utilizei o 'from None' para preservar o traceback
                    raise ModuleNotFoundError(msg)
            except ImportError as e:
                msg = f"Modelo: {model_name}. Erro ao importar os módulos necessários para o módulo " \
                      f"'{caminho_import}'. Mensagem do Import: {str(e)}"
                LOGGER.error(msg)
                # Não utilizei o 'from None' para preservar o traceback
                raise ImportError(msg)

            if models_params[model_name]['model_class'] == "ModeloCLF":
                cls = getattr(modulo, "ModeloCLF")

                if callable(cls) and type(cls).__name__ == 'ABCMeta':
                    try:
                        modelos[model_name] = cls(model_name=model_name,
                                                  model_provider_name=models_params[model_name]['model_provider_name'])
                    except TypeError as e:
                        msg = f"Faltou a implementação do(s) seguinte(s) método(s) para o modelo '{model_name}' " \
                              f"(classe 'ModeloCLF'): {str(e)[64:]}."
                        LOGGER.error(msg)
                        raise TypeError(msg) from None
                else:
                    msg = f"Modelo: {model_name}. O tipo do 'ModeloCLF' está incorreto: '{type(cls).__name__}'. " \
                          f"'ModeloCLF' deve ser uma classe que herda os métodos da interface " \
                          f"'ModelPublicationInterfaceCLF' e possua as implementações para os métodos abstratos dela."
                    LOGGER.error(msg)
                    raise TypeError(msg)
            elif models_params[model_name]['model_class'] == "ModeloRETRAIN":
                cls = getattr(modulo, "ModeloRETRAIN")

                if callable(cls) and type(cls).__name__ == 'ABCMeta':
                    try:
                        modelos[model_name] = cls(model_name=model_name,
                                                  model_provider_name=models_params[model_name]['model_provider_name'],
                                                  experiment_name=models_params[model_name]['experiment_name'],
                                                  dataset_provider_name=models_params[model_name]['dataset_provider_name'])
                    except TypeError as e:
                        msg = f"Faltou a implementação do(s) seguinte(s) método(s) para o modelo '{model_name}' " \
                              f"(classe 'ModeloRETRAIN'): {str(e)[68:]}."
                        LOGGER.error(msg)
                        raise TypeError(msg) from None
                else:
                    msg = f"Modelo: {model_name}. O tipo do 'ModeloRETRAIN' está incorreto: '{type(cls).__name__}'. " \
                          f"'ModeloRETRAIN' deve ser uma classe que herda os métodos da interface " \
                          f"'ModelPublicationInterfaceRETRAIN' e possua as implementações para os métodos " \
                          f"abstratos dela."
                    LOGGER.error(msg)
                    raise TypeError(msg)
            else:
                msg = f"Modelo: {model_name}. O valor do parâmetro 'model_class' está incorreto. Foi informado " \
                      f"'{models_params[model_name]['model_class']}' no arquivo 'params.conf', porém deve ser " \
                      f"'ModeloCLF' ou 'ModeloRETRAIN'. Programa abortado!"
                LOGGER.error(msg)
                raise ValueError(msg)

        return modelos



# ----------------------------------------------------------------------------------------------------
# Provider para obtenção de datasets persistidos no Minio
# ----------------------------------------------------------------------------------------------------
import os
from io import BytesIO
from ..utils import load_env_variables, get_file_s3, make_log

# Para facilitar, define um logger único para todas as funções
LOGGER = make_log("LOG_MLLIB.log")


def load_datasets_minio(datasets_filenames: dict) -> dict:
    """
    Carrega os datasets necessários para o modelo que foram persistidos utilizando o Minio. Os parâmetros de acesso
    deverão ser fornecidos através de um arquivo chamado '.env' que deve ser criado no repositório local e preenchido
    com as seguintes variáveis: MINIO = "nome do servidor s3", ACCESS_KEY = "chave de acesso", SECRET_KEY = "senha de
    acesso" e BUCKET = "nome do bucket". Dica de segurança: Não deixe o arquivo '.env' ser versionado/persistido no
    repositório remoto do código.
        :param datasets_filenames: Dicionário contendo os tipos de datasets e os nomes dos respectivos arquivos.
                                   Exemplo: {'features': 'nome_arquivo_features', 'targets': 'nome_arquivo_targets'}
        :return: Dicionário com os datasets carregados.
    """
    # Obtém as credenciais e as informações necessárias para baixar os arquivos
    load_env_variables()
    s3_server = os.environ.get("MINIO")
    access_key = os.environ.get("ACCESS_KEY")
    secret_key = os.environ.get("SECRET_KEY")
    bucket = os.environ.get("BUCKET")

    if s3_server is None or access_key is None or secret_key is None or bucket is None:
        msg = f"Não foram encontradas todas as variáveis de ambiente necessárias. Certifique-se que um arquivo " \
              f"chamado '.env' exista; esteja localizado na pasta da aplicação e que possua valores para as " \
              f"variáveis: 'MINIO', 'ACCESS_KEY', 'SECRET_KEY' e 'BUCKET'. Ou se preferir, configure essas variáveis " \
              f"de ambiente e seus respectivos valores. Programa abortado!"
        LOGGER.error(msg)
        raise RuntimeError(msg)

    datasets = {}

    for tipo, nome_arquivo in datasets_filenames.items():
        datasets[tipo] = BytesIO(get_file_s3(nome_arquivo, s3_server, access_key, secret_key, bucket))

    return datasets
